#!/usr/bin/env python
# -*- coding: iso-8859-15 -*-

# GIMP script for several stroke-editing functions
# (c) Ofnuts 2020
#
#   History:
#
#   v0.00: 2016-12-03 Gather separate scripts, refactor code
#   v0.01: 2016-12-11 Fix Extract/Delete bug when path extends beyond canvas limits
#   v0.02: 2019-02-04 Add "Break text path part"
#   v0.03: 2019-02-05 Set visibility of results in "Break path apart" and "Break text path part"
#   v0.04: 2019-06-30 Add sort horizontally, vertically
#   v0.05: 2020-01-07 Add "Extract/Delete Anchors"
#   v0.06: 2020-05-09 Add stroke direction in the summary
#   v0.07: 2020-05-21 Much better algorithm for "Break text path apart"
#   v0.08: 2020-06-21 Add "Densify anchors"
#   v0.09: 2020-08-27 Add "Break path apart by nesting order"
#   v0.10: 2021-03-30 Use unicode symbols for direction
#                     CPU & RAM Performance improvements
#          2022-05-05 added German translation @gimplyworxs gimp-werkstatt.de
# 					  please remove any previously installed version of the plug-in


#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published
#   by the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This very file is the complete source code to the program.
#
#   If you make and redistribute changes to this code, please mark it
#   in reasonable ways as different from the original version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   The GPL v3 licence is available at: https://www.gnu.org/licenses/gpl-3.0.en.html


import os, sys, math
from collections import namedtuple
from gimpfu import *
import traceback

debug='OFN_DEBUG' in os.environ

def trace(format,*args):
    if debug:
        print format % args

# ***********************************************
# Some useful functions and classes

# *********
# Point class, that derives from namedtuple. This makes it immutable, and behaving as an
# iterable list with x,y which is useful when passing x and y as parameters ("*point").
class Point(namedtuple('Point', 'x y')):
    __slots__ = () # Save memory since we will potentially allocate a lot of these small objects

    CLOSE_ENOUGH=0.1

    def __repr__(self):
        return '{%4.2f,%4.2f}' % (self.x, self.y)

    def __eq__(self,other):
        return isinstance(other,Point) and abs(self.x-other.x) < self.CLOSE_ENOUGH and abs(self.y-other.y) < self.CLOSE_ENOUGH

    @property
    def pixelX(self):
        return int(math.floor(self.x))

    @property
    def pixelY(self):
        return int(math.floor(self.y))

    @property
    def pixel(self):
        return int(math.floor(self.x)),int(math.floor(self.y))

# *********
# Utility functions

class CircularList(object):
    '''
    Extends list with indices and slices that can wrap around
    '''
    def __init__(self,items):
        self._items=items

    def __getitem__(self,i):
        n=len(self._items)
        if isinstance(i,slice):
            if i.start and i.stop:
                return [self._items[j%n] for j in xrange(*i.indices(2*n))]
            elif i.start and not i.stop:
                return [self._items[j%n] for j in xrange(*slice(i.start,i.start+n,i.step).indices(2*n))]
            else:
                return self._items[i]
        elif isinstance(i,int):
                return self._items[i%n]
        else:
            return self._items[i] # Likely trigger standard error

    def __iter__(self):
        return self._items.__iter__()

    def __len__(self):
        return len(self._items)

    def pairs(self):
        for i in range(len(self._items)):
            yield self[i:i+2]

    @property
    def items(self):
        return self._items

def pointsFromCoords(coords):
    '''Transforms a flat list of coordinates into a list of Points'''
    uniqueiterator=iter(coords)
    return [Point(x,y) for x,y in zip(uniqueiterator,uniqueiterator)]

def coordsFromPoints(points):
    '''Flattens a list of Points into a list of coordinates'''
    return [coord for point in points for coord in point]

def xCoords(points):
    return [p.x for p in points]

def yCoords(points):
    return [p.y for p in points]

def anchorIterator(points):
    '''Iterates the anchors in a list of Points'''
    for i in range(1,len(points),3):
        yield points[i]

def boundingBox(points):
    allX=[p.x for p in points]
    allY=[p.y for p in points]
    return min(allX),min(allY),max(allX),max(allY)

def boxInsideBox(inner,outer):
    innerMinX,innerMinY,innerMaxX,innerMaxY=inner
    outerMinX,outerMinY,outerMaxX,outerMaxY=outer
    return innerMinX>=outerMinX and innerMaxX<=outerMaxX and innerMinY>=outerMinY and innerMaxY<=outerMaxY

# Determines if point inside polygon using the ray-tracing algorithm
# If an horizontal to the right of the point intersects an odd number
# of segments the point is inside
def pointInsidePolygon(p,polygon):
    inside = False
    x,y=p
    for p1,p2 in polygon.pairs():
        p1x,p1y=p1
        p2x,p2y=p2
        if y > min(p1y,p2y) and y <= max(p1y,p2y) and x <= max(p1x,p2x):
            if p1y != p2y:
                xints = (y-p1y)*(p2x-p1x)/(p2y-p1y)+p1x
            if p1x == p2x or x <= xints:
                inside = not inside
    return inside

# Polygon A is inside polygon B if all points of A are inside B
# There are edge cases where this is wrong and we also could/should check
# that no point of B is inside A... (should be very rare in practice)
def polygonInsidePolygon(inner,outer):
    for p in inner:
        if not pointInsidePolygon(p,outer):
            return False
    return True

# Mid point of Bezier spline
def midPoint(points):
    def midCoord(a,b,c,d):
        return (a+3*b+3*c+d)/8
    return Point(midCoord(*xCoords(points)),midCoord(*yCoords(points)))

# Returns the anchors and tangents to add an anchor midway between two anchors
def densifyCoords(coords):
    a,b,c,d=coords
    return [a,(a+b)/2,(a+2*b+c)/4,(a+3*b+3*c+d)/8,(b+2*c+d)/4,(c+d)/2,d]

def densifySpline(points):
    return [Point(x,y) for x,y in zip(densifyCoords(xCoords(points)),densifyCoords(yCoords(points)))]

# Envelope polygon from stroke using anchors and
# additional midpoints between anchors for enhanced accuracy
def polygon(strokePoints):
    loPrez=CircularList(strokePoints)
    n=len(loPrez)
    hiPrez=[(loPrez[i],midPoint(loPrez[i:i+4])) for i in range(1,n,3)]
    return CircularList([p for pair in hiPrez for p in pair])

# ******************
# Stroke class, which helps handling the list of points.
# Unlike gimp.VectorBezierStroke it can exist outside of a path.
class Stroke(object):
    '''
    Stroke class, which helps handling the list of points.
    Unlike gimp.VectorBezierStroke it can exist outside of a path.
    '''
    ids=0
    # General constructor, accepts gimp.Stroke, flat list of coordinates, or Points
    def __init__(self,data,closed=False):
        self._closed=closed
        if isinstance(data,gimp.VectorsBezierStroke):
            coords,self._closed=data.points
            self._points=pointsFromCoords(coords)
            self._id=str(data.ID)
        elif isinstance(data,list):
            if isinstance(data[0],Point):
                self._points=data
            elif isinstance(data[0],float):
                self._points=pointsFromCoords(coords)
            else:
                raise Exception('Cannot create stroke from list of %s' % type(data[0]))
            self.ids+=1
            self._id=str(self.ids)+'?'
        else:
            raise Exception('Cannot create stroke from %s' % type(data))

        self._polygon=polygon(self._points)
        self._boundingBox=boundingBox(self._points)
        self.children=[]

    def __repr__(self):
        return '%r<-%s,[%d]->%r' % (self.start, 'OC'[self._closed], len(self.points),self.end)

    # create reversed stroke
    def reversed(self):
        return Stroke(self._points[::-1],self._closed)

    def gimpStroke(self,path):
        return gimp.VectorsBezierStroke(path,coordsFromPoints(self._points),self._closed)

    # Some fake read-only attributes
    @property
    def id(self):
        return self._id

    @property
    def points(self):
        return self._points

    @property
    def closed(self):
        return self._closed

    # Some specifc often-used points
    @property
    def start(self):
        return self._points[1]

    @property
    def end(self):
        return self._points[-2]

    @property
    def firstForward(self):
        return self._points[2]

    @property
    def lastBackward(self):
        return self._points[-3]

    @property
    def boundingBox(self):
        return self._boundingBox

    @property
    def center(self):
        minX,minY,maxX,maxY=self._boundingBox
        return (minX+maxX)/2,(minY+maxY)/2

    @property
    def centerX(self):
        minX,_,maxX,_=self._boundingBox
        return (minX+maxX)/2

    @property
    def centerY(self):
        _,minY,_,maxY=self._boundingBox
        return (minY+maxY)/2

    @property
    def polygon(self):
        return self._polygon

    def contains(self,other):
        # Fast check then slow check when necessary
        return boxInsideBox(other.boundingBox,self.boundingBox) and polygonInsidePolygon(other.polygon,self.polygon)

    def dumpTree(self,level):
        print '%s %s' % ('  '*level,self.id)
        for s in self.children:
            s.dumpTree(level+1)

##############################################
# Making many changes to a path appears to be costly because Gimp creates an undo level
# for each change (added/removed stroke). So instead of altering paths in place we make
# a copy and modify it before it is attached to the image, so the only undo steps are the
# two implied by the removal of the original and the insertion of the copy.

def copyPath(path,withStrokes):
    newPath=gimp.Vectors(path.image,"Copy of %s" % path.name)
    if withStrokes:
        for s in path.strokes:
            points,closed=s.points
            gimp.VectorsBezierStroke(newPath,points,closed)
    return newPath

def replacePath(oldPath,newPath):
    image=oldPath.image
    position=pdb.gimp_image_get_item_position(image,oldPath)
    name=oldPath.name

    newPath.linked=oldPath.linked
    newPath.tattoo=oldPath.tattoo
    newPath.visible=oldPath.visible
    pdb.gimp_image_remove_vectors(image, oldPath)
    newPath.name=name
    pdb.gimp_image_insert_vectors(image, newPath, None, position)
    return newPath

#######################################################

def isSelectedPoint(point,image):
    x,y=point.pixelX,point.pixelY
    return 0<=x<image.width and 0<=y<image.height and image.selection.get_pixel(x,y)[0] > 0

def hasSelectedAnchors(image,points,includeEnds=True):
    '''Returns first anchor in selection, if any'''
    if includeEnds:
        checkedAnchors=enumerate(anchorIterator(points),start=0)
    else:
        checkedAnchors=enumerate(anchorIterator(points[3:-3]),start=1)
    for i,point in checkedAnchors:
        if isSelectedPoint(point,image):
            return True,i
    return False,-10000 # Make sure it is not used by mistake

# **************************
# Delete and extract anchors

def deleteOrExtractAnchors(image,path,extract):
    if pdb.gimp_selection_is_empty(image):
        raise Exception('Empty selection, nothing to %s' % ('extract' if extract else 'delete'))

    newPath=copyPath(path,False)

    for stroke in path.strokes:
        coords,closed=stroke.points
        points=pointsFromCoords(coords)
        remainingPoints=[]
        for i in range(0,len(points),3):
            triplet=points[i:i+3]
            if isSelectedPoint(triplet[1],image)^(not extract):
                remainingPoints.extend(triplet)
        if remainingPoints:
            Stroke(remainingPoints,closed).gimpStroke(newPath)

    replacePath(path,newPath)

def deleteAnchors(image,sourcePath):
    deleteOrExtractAnchors(image,sourcePath,False)

def extractAnchors(image,sourcePath):
    deleteOrExtractAnchors(image,sourcePath,True)


# **************************
# Delete and extract strokes

def deleteOrExtractStrokes(image,path,extract):
    if pdb.gimp_selection_is_empty(image):
        raise Exception('Empty selection, nothing to %s' % ('extract' if extract else 'delete'))

    newPath=copyPath(path,False)
    for stroke in path.strokes:
        coords,closed=stroke.points
        selected,_=hasSelectedAnchors(image,pointsFromCoords(coords))
        if not selected^extract:
            Stroke(stroke).gimpStroke(newPath)
    replacePath(path,newPath)

def deleteStrokes(image,sourcePath):
    deleteOrExtractStrokes(image,sourcePath,False)

def extractStrokes(image,sourcePath):
    deleteOrExtractStrokes(image,sourcePath,True)

# *************************
# Reverse Strokes

def reverseStrokes(image,path):
    reverseAll=pdb.gimp_selection_is_empty(image)
    newPath=copyPath(path,False)

    # To keep strokes in order we have to remove them all
    # And add them back with changes
    strokes=[Stroke(stroke) for stroke in path.strokes]
    for stroke in strokes:
        selected,_=hasSelectedAnchors(image,stroke.points)
        if reverseAll or selected:
            stroke.reversed().gimpStroke(newPath)
        # add back
        else:
            stroke.gimpStroke(newPath)
    replacePath(path,newPath)

# *************************
# Join strokes

# Attempts to properly close a Stroke if begin/end points are close to each other
# If so, the end point is removed, its backward tangent becomes the backward tangent
# of the first point.
# If closure is not possible or already done, the original Stroke is returned.
def attemptClose(s):
    if s.closed:
        return s
    elif s.start==s.end:
        #print 'Closing %s' % s
        # If no backward tangent on removed point, move handle to remaining point
        firstTangent= s.start if s.lastBackward==s.end else s.lastBackward
        return Stroke([firstTangent]+s.points[1:-3],closed=True)
    else:
        return s

# Attempts to join two strokes.
# If join is not possible, returns the first stroke.
def attemptJoin(s1,s2):
    #print 'Joining %s and %s' % (s1,s2)
    if s1.closed or s2.closed:
        return s1
    elif s1.start==s2.start:
        s1,s2=s1.reversed(),s2
    elif s1.start==s2.end:
        s1,s2=s2,s1
    elif s1.end==s2.start:
        pass; # s1,s2 already set
    elif s1.end==s2.end:
        s1,s2=s1,s2.reversed()
    else:
        return s1
    # Now perform actual join, dropping the forward handle of the last anchor
    # of s1 and the first anchor of s2 and its forward handle.
    # If there is no backward handle on the first anchor of s2, use the coordinates
    # of the last anchor of s1 instead.
    tangent=s1.end if s2.firstForward==s2.end else s2.firstForward
    return Stroke(s1.points[:-1]+[tangent]+s2.points[3:],closed=False)

def joinStrokes(image,path):
    removeStrokes=path.strokes[:]
    sourceStrokes=[Stroke(gimpStroke) for gimpStroke in path.strokes]
    finalStrokes=[]

    while sourceStrokes:
        stroke=sourceStrokes.pop()
        #print 'Stroke: %s' % stroke
        # if closable or closed no need to look for other strokes
        stroke=attemptClose(stroke)
        if stroke.closed:
            finalStrokes.append(stroke)
            continue # Nothing more to do with that one
        # Scan remaining strokes for possible joins
        for candidate in sourceStrokes:
            joinedStroke=attemptJoin(stroke,candidate)
            if joinedStroke is stroke: # no splice occured
                continue
            else:
                #print 'Joined: %s' % joinedStroke
                sourceStrokes.remove(candidate) # No longer exists
                sourceStrokes.append(joinedStroke) # recheck for closure or more joins
                break
        else: # executed only if we didn't join anything
            finalStrokes.append(stroke)

    # make new path with joined strokes
    newPath=copyPath(path,False)
    for stroke in finalStrokes:
        stroke.gimpStroke(newPath)
    replacePath(path,newPath)

# *************************
# Disjoin Stroke

def disjoinOpenStrokeOnAnchor(stroke,anchor):
    '''Returns two open strokes, starting/ending on given anchor'''
    points1=stroke.points[:(anchor+1)*3]
    points2=stroke.points[anchor*3:]
    return (Stroke(points1,False),Stroke(points2,False))

def disjoinClosedStrokeOnAnchor(stroke,anchor):
    '''Returns one single open stroke, starting and ending on given anchor (begin and end overlap)
       This stroke has of course one more anchor than the original one'''
    ring=stroke.points+stroke.points
    pointsCount=len(stroke.points)+3 # Number of points to extract from ring
    # return a single-element tuple
    return (Stroke(ring[anchor*3:anchor*3+pointsCount],False),)

def disjoinStrokeOnAnchor(stroke,anchor):
    if stroke.closed:
        return disjoinClosedStrokeOnAnchor(stroke,anchor)
    else:
        return disjoinOpenStrokeOnAnchor(stroke,anchor)

def disjoinStrokes(image,path):
    removeStrokes=path.strokes[:]
    sourceStrokes=[Stroke(gimpStroke) for gimpStroke in path.strokes]
    finalStrokes=[]

    while sourceStrokes:
        candidate=sourceStrokes.pop()
        selected,anchor=hasSelectedAnchors(image,candidate.points,candidate.closed)
        if selected and len(candidate.points)>1:
            disjoinStrokes=disjoinStrokeOnAnchor(candidate,anchor)
            sourceStrokes.extend(disjoinStrokes)
        else:
            finalStrokes.append(candidate)

    # make new path with final strokes
    newPath=copyPath(path,False)
    for stroke in finalStrokes:
        stroke.gimpStroke(newPath)
    replacePath(path,newPath)

# *************************
# Sort strokes
# Strokes are removed and then reinserted in order
def sortStrokesHorizontally(image,path):
    newPath=copyPath(path,False)
    strokes=[Stroke(stroke) for stroke in path.strokes]
    strokes.sort(key=lambda s : s.centerX)
    for stroke in strokes:
        stroke.gimpStroke(newPath)
    replacePath(path,newPath)

def sortStrokesVertically(image,path):
    newPath=copyPath(path,False)
    strokes=[Stroke(stroke) for stroke in path.strokes]
    strokes.sort(key=lambda s : s.centerY)
    for stroke in strokes:
        stroke.gimpStroke(newPath)
    replacePath(path,newPath)

# *************************
# Densify anchors
def densifyStroke(s):
    if s.closed:
        # Add extra points so we can densify the "virtual closure" spline
        points=s.points[:]+s.points[0:3]
    else:
        points=s.points
    densifiedPoints=[]
    for i in xrange(1,len(points)-3,3):
        spline=points[i:i+4]
        densifiedPoints+=densifySpline(spline)[:-1]

    if s.closed:
        # Add first backard tangent (the backward tangent of the last densified point)
        # And drop last tangent and anchor (virtual closure spline)
        final=densifiedPoints[-1:]+densifiedPoints[:-1]
    else:
        # add first backard tangent, final point, and final forward tangent
        final=points[0:1]+densifiedPoints+points[-2:]

    return Stroke(final,s.closed)

def densifyPath(image,path):
    newPath=copyPath(path,False)
    sourceStrokes=path.strokes[:]
    for s in sourceStrokes:
        sourceStroke=Stroke(s)
        densifiedStroke=densifyStroke(sourceStroke)
        densifiedStroke.gimpStroke(newPath)
    replacePath(path,newPath)

# *************************
# Break path apart
def breakPathApart(image,path):
    if path==None:
        raise Exception('No active path in image')
    if not path :
        raise Exception('No elements in active path')

    for stroke in path.strokes:
        strokePath=gimp.Vectors(image,'%s [%d]' % (path.name,stroke.ID))
        points,closed=stroke.points
        gimp.VectorsBezierStroke(strokePath,points,closed)
        pdb.gimp_image_insert_vectors(image, strokePath, None,-1)
        strokePath.visible=True
    path.visible=False

# *************************
# Make a nesting tree with the strokes
def makeNestingTree(strokes):
    topLevel=[]
    for s in strokes:
        for other in strokes:
            if s is other:
                continue
            if other.contains(s):
                other.children.append(s)
                break
        else:
            topLevel.append(s)
    for s in topLevel:
        s.children=makeNestingTree(s.children)
    return topLevel

# *************************
# Break text path apart: keep together nested strokes
def addStrokeToPath(stroke,path):
    stroke.gimpStroke(path)
    for nestedStroke in stroke.children:
        addStrokeToPath(nestedStroke,path)

def breakTextPathApart(image,path):
    if path==None:
        raise Exception('No active path in image')
    if not path :
        raise Exception('No elements in active path')

    strokes=[Stroke(s) for s in path.strokes if s.points[1]] # All closed strokes

    # Create a tree view of the Strokes, based on nesting:
    # Top level: outlines, 2nd level: holes, 3rd: outlines in holes, etc...
    topLevelStrokes=makeNestingTree(strokes)

    # Create a path for each group of nested strokes
    for i,stroke in enumerate(topLevelStrokes,1):
        letterPath=letterPath=gimp.Vectors(image,'%s [%d]' % (path.name,i))
        addStrokeToPath(stroke,letterPath)
        pdb.gimp_image_insert_vectors(image, letterPath, None,-1)
        letterPath.visible=True
    path.visible=False

# *************************
# Break path apart nested: make a path with all the strokes at the same nesting level
def addToLevel(stroke,strokeLevels,level):
    while len(strokeLevels)<=level:
        strokeLevels.append([])
    strokeLevels[level].append(stroke)
    for nestedStroke in stroke.children:
        addToLevel(nestedStroke,strokeLevels,level+1)

def breakPathApartNested(image,path):
    if path==None:
        raise Exception('No active path in image')
    if not path :
        raise Exception('No elements in active path')

    strokes=[Stroke(s) for s in path.strokes if s.points[1]] # All closed strokes

    # Create a tree view of the Strokes, based on nesting:
    # Top level: outlines, 2nd level: holes, 3rd: outlines in holes, etc...
    topLevelStrokes=makeNestingTree(strokes)

    # Make lists of strokes by nesting level
    strokeLevels=[]
    for stroke in topLevelStrokes:
        addToLevel(stroke,strokeLevels,0)

    # Create a path for each level of strokes
    for i,strokeLevel in enumerate(strokeLevels,1):
        levelPath=gimp.Vectors(image,'%s[%d]' % (path.name,i))
        for stroke in strokeLevel:
            stroke.gimpStroke(levelPath)
        pdb.gimp_image_insert_vectors(image,levelPath, None,0)
        levelPath.visible=True

    path.visible=False

# *************************
# Path summary
def strokeSummary(path,stroke):
    # Compute orientation, see https://en.wikipedia.org/wiki/Curve_orientation
    def strokeType(stroke):
        coords,closed=stroke.points
        if not closed:
            return 0
        points=pointsFromCoords(coords)
        anchors=points[1::3]
        vertex=Point(sys.maxint,sys.maxint)
        index=-1
        for i in range(len(anchors)):
            if anchors[i]<vertex:
                vertex=anchors[i]
                index=i
        a,b,c=(anchors[-1:]+anchors+anchors[0:1])[index:index+3]
        det=(b.x-a.x)*(c.y-a.y)-(c.x-a.x)*(b.y-a.y)
        return 1 if det>=0. else -1

    points,closed=stroke.points
    length=pdb.gimp_vectors_stroke_get_length(path, stroke.ID, .1)
    return '(%3.1f, %3.1f) <- %d : %3.1f : %s -> (%3.1f, %3.1f)' % \
        ( points[2],points[3],
         len(points)/6,length,
         ['↺','↝','↻'][1+strokeType(stroke)],
         points[-4],points[-3])

def pathSummary(image,path):
    if path==None:
        raise Exception('No active path in image')
    if not path :
        raise Exception('No elements in active path')

    s=[Stroke(stroke) for stroke in path.strokes]

    strokeSummaries=[strokeSummary(path,stroke) for stroke in path.strokes]
    pdb.gimp_message('Path "%s" [%d]:\n%s' % (path.name,len(strokeSummaries),'\n'.join(strokeSummaries)))

# *************************************************************************************
# Registrations

whoiam='\n'+os.path.abspath(sys.argv[0])
author='Ofnuts'
year='2020'
menu='<Vectors>/Edit'

def protected(function):
    '''Create protected version of function (general try/catch, and Gimp undo)'''
    def p(*parms):
        image=parms[0]
        pdb.gimp_image_undo_group_start(image)
        'gimp_image_freeze_vectors' in dir(pdb) and pdb.gimp_image_freeze_vectors(image)
        try:
            function(*parms)
        except Exception as e:
            print e.args[0]
            traceback.print_exc()
            pdb.gimp_message(e.args[0])
        'gimp_image_freeze_vectors' in dir(pdb) and pdb.gimp_image_thaw_vectors(image)
        pdb.gimp_image_undo_group_end(image)
        pdb.gimp_displays_flush()
    return p

def commonRegister(gimpProc,pythonProc,desc):
    '''Common registration function'''
    register(
        'ofn-path-'+gimpProc,desc+'...'+whoiam,desc,author,author,year,desc,'*',
        [(PF_IMAGE, 'image', 'Input image', None),(PF_VECTORS, 'refpath', 'Input path', None)],[],
        protected(pythonProc),menu=menu
    )

# commonRegister('delete-anchors',deleteAnchors,'Delete anchors')
# commonRegister('extract-anchors',extractAnchors,'Extract anchors')
# commonRegister('delete-strokes',deleteStrokes,'Delete strokes')
# commonRegister('extract-strokes',extractStrokes,'Extract strokes')
# commonRegister('reverse-strokes',reverseStrokes,'Reverse strokes')
# commonRegister('join-strokes',joinStrokes,'Join strokes')
# commonRegister('sort-strokes-horizontally',sortStrokesHorizontally,'Sort strokes horizontally')
# commonRegister('sort-strokes-vertically',sortStrokesVertically,'Sort strokes vertically')
# commonRegister('disjoin-strokes',disjoinStrokes,'Disjoin strokes')
# commonRegister('break-path-apart',breakPathApart,'Break path apart')
# commonRegister('break-text-path-apart',breakTextPathApart,'Break text path apart')
# commonRegister('break-path-apart-nested',breakPathApartNested,'Break path apart by nesting order')
# commonRegister('path-summary',pathSummary,'Path summary')
# commonRegister('densify-anchors',densifyPath,'Densify anchors')

commonRegister('delete-anchors',deleteAnchors,'Anker löschen')
commonRegister('extract-anchors',extractAnchors,'Anker extrahieren')
commonRegister('delete-strokes',deleteStrokes,'Pfade löschen')
commonRegister('extract-strokes',extractStrokes,'Pfade extrahieren')
commonRegister('reverse-strokes',reverseStrokes,'Pfade umkehren')
commonRegister('join-strokes',joinStrokes,'Pfade zusammenfügen')
commonRegister('sort-strokes-horizontally',sortStrokesHorizontally,'Pfade horizontal sortieren')
commonRegister('sort-strokes-vertically',sortStrokesVertically,'Pfade vertikal sortieren')
commonRegister('disjoin-strokes',disjoinStrokes,'Pfade entkoppeln')
commonRegister('break-path-apart',breakPathApart,'Pfad aufbrechen')
commonRegister('break-text-path-apart',breakTextPathApart,'Text Pfad aufbrechen')
commonRegister('break-path-apart-nested',breakPathApartNested,'Pfad nach Verschachtelungsreihenfolge zerlegen')
commonRegister('path-summary',pathSummary,'Pfadzusammenfassung')
commonRegister('densify-anchors',densifyPath,'Anker verdichten')

main()
